import configparser
import os
from pathlib import Path
from typing import Union, List


class Config:
    """Class for working with the configuration file."""

    _default_deck = 'Default'
    _default_folder = ''
    _default_profile = ''
    _default_basic_type = 'Basic'
    _default_front_field = 'Front'
    _default_back_field = 'Back'
    _default_cloze_type = 'Cloze'
    _default_cloze_field = 'Text'
    _default_port = '8765'
    _default_highlight_style = 'monokai'

    def __init__(self, config_path: Union[str, Path]):
        self._config = configparser.ConfigParser()
        self._config_path = config_path

        if not os.path.exists(self._config_path):
            self._create_default()
        else:
            self._read()

    def reset(self):
        """Reset config file to default state"""
        self._config = configparser.ConfigParser()
        self._create_default()

    def _create_default(self):
        """Create default configuration file"""
        config_dict = {
            'defaults': {
                'profile': self._default_profile,
                'deck': self._default_deck,
                'folder': self._default_folder,
            },
            'anki': {
                'basic_type': self._default_basic_type,
                'front_field': self._default_front_field,
                'back_field': self._default_back_field,
                'cloze_type': self._default_cloze_type,
                'cloze_field': self._default_cloze_field
            },
            'anki_connect': {
                'port': self._default_port
            },
            'highlight': {
                'style': self._default_highlight_style,
            },
        }

        self._config.read_dict(config_dict)
        self._save()

    def _read(self):
        """Get config from file system"""
        self._config.read(self._config_path)

    def _save(self):
        """Save config state in file system"""
        with open(self._config_path, mode='wt', encoding='utf-8') as file:
            self._config.write(file)

    def get_option_value(self, section: str, key: str) -> str:
        """Get value of the config entry"""
        return self._config[section][key]

    def update_option_value(self, section: str, key: str, new_value: str):
        """Update value of the config entry"""
        if key not in self._config[section]:
            raise KeyError

        self._config[section][key] = new_value
        self._save()

    def get_formatted_options(self) -> List[str]:
        """Get list of formatted key-value entries from the config"""
        formatted_entries = []
        for section in self._config.sections():
            for key, value in self._config[section].items():
                formatted_entries.append(f'{section}.{key} = {value}')

        return formatted_entries

    def __repr__(self):
        return f"{type(self).__name__}(config_path={repr(self._config_path)})"



def create_anki_search_query(text: str) -> str:
    """Create Anki search query from the supplied text"""
    special_chars = ['\\', '"', '*', '_', ':']

    search_query = text
    for char in special_chars:
        escaped_char = '\\' + char
        search_query = search_query.replace(char, escaped_char)

    return '"' + search_query + '"'



from pathlib import Path
from typing import Union, List, Optional

from .notes.basic_note import BasicNote
from .notes.cloze_note import ClozeNote
from .parser import Parser


class Writer:
    """Class for editing file with notes"""

    def __init__(self, file_path: Union[str, Path], notes: List[Union[BasicNote, ClozeNote]]):
        self._file_path = file_path
        self._notes = notes

        with open(self._file_path, mode='rt', encoding='utf-8') as f:
            self._file_content = f.read()
        self._note_strings = Parser.get_note_strings(self._file_content)

    def update_note_ids(self):
        """Update lines with IDs of the notes from the file"""
        for note in self._notes:
            # Find note's question in file string
            note_question = note.get_raw_question_field()
            question_start = self._file_content.find(note_question)

            # Note may not be found cause it was deleted
            if question_start == -1:
                continue

            # Find newline character that comes before question
            newline_index = self._file_content[:question_start].rfind('\n')

            # Search for existing line with ID
            line_before_question = self._file_content[:newline_index].splitlines()[-1]
            existing_id = Parser.get_id(line_before_question)

            # Skip if ID hasn't changed
            if existing_id == note.anki_id:
                continue

            id_string = f'<!--ID:{note.anki_id}-->\n'
            if not note.anki_id:
                # Delete ID from note in file if note object has no ID
                id_string = ''

            if existing_id is None:
                # Add line with ID after newline
                self._file_content = (self._file_content[:newline_index + 1] +
                                      id_string +
                                      self._file_content[newline_index + 1:])
            else:
                # Substitute existing ID with the new one
                self._file_content = self._file_content.replace(f'<!--ID:{existing_id}-->\n', id_string)

        self._save()

    def update_fields_of_basic_notes(self):
        """Update question and answer fields in notes in file"""
        for note in self._notes:
            if not note.changed:
                continue

            # Find string with this note by its ID
            note_string = self._get_note_string_by_id(note.anki_id)

            # Substitute question field
            current_question = Parser.get_question(note_string)
            self._file_content = self._file_content.replace(current_question, note.raw_front_md)

            # Create new answer field (with '>')
            current_answer = Parser.get_answer(note_string).rstrip()
            lines = note.raw_back_md.replace('\n\n', '\n').splitlines()
            new_answer = '\n'.join(map(lambda line: f'> {line}', lines))

            # Substitute answer field
            self._file_content = self._file_content.replace(current_answer, new_answer)

        self._save()

    def delete_notes(self):
        """Delete notes marked for deletion from the file"""
        for note in self._notes:
            if not note.to_delete:
                continue

            # Find string with this note by its ID
            note_string = self._get_note_string_by_id(note.anki_id)

            # Delete note text from file
            self._file_content = self._file_content.replace(note_string, '')

        self._save()

    def update_cloze_notes(self):
        """Updates all cloze notes with the values from updated_text_md field"""
        for note in self._notes:
            if not isinstance(note, ClozeNote):
                continue

            self._file_content = self._file_content.replace(note.raw_text_md, note.updated_text_md, 1)
            note.raw_text_md = note.updated_text_md  # update info about raw text

        self._save()

    def _save(self):
        """Save file state into the file system"""
        with open(self._file_path, mode='wt', encoding='utf-8') as f:
            f.write(self._file_content)
        self._note_strings = Parser.get_note_strings(self._file_content)

    def _get_note_string_by_id(self, note_id: int) -> Optional[str]:
        """Gets note string from the file by its ID. If note wasn't found returns None"""
        for string in self._note_strings:
            if string.find(str(note_id)) != -1:
                return string

    def __repr__(self):
        return f'{type(self).__name__}(file_path={self._file_path!r}, notes={self._notes!r})'



import re
from pathlib import Path
from typing import List, Union, Optional

from .notes.basic_note import BasicNote
from .notes.cloze_note import ClozeNote


class Parser:
    """Class for getting notes and various information about them from the text file"""

    _section_regex = re.compile(
        r'^---\n'  # start of section
        r'(.+?)'  # contents
        r'^---$',  # end of section
        re.MULTILINE | re.DOTALL
    )
    _deck_name_regex = re.compile(
        r'(?<=^Deck:)(.*?)$',
        re.MULTILINE
    )
    _tags_regex = re.compile(
        r'(?<=^Tags:)(.*?)$',
        re.MULTILINE
    )
    _all_notes_regex = re.compile(
        r'(?:^<!--ID:\S+-->\n)?'  # optional ID
        r'^\d+\.[\s\S]*?'  # card contents (optionally contains answer)
        r'(?=<!--ID:\S+-->|^\d+\.|\Z|^---$)',  # card ends before ID, start of next card, end of string, end of section
        re.MULTILINE
    )
    _basic_note_regex = re.compile(
        r'(?:^<!--ID:\S+-->\n)?'  # optional ID
        r'^\d+\.[\s\S]+?'  # card question
        r'(?:^>.*?(?:\n|$))+',  # card answer
        re.MULTILINE
    )
    _cloze_note_regex = re.compile(
        r"(?:^<!--ID:\S+-->\n)?"  # optional ID
        r"^\d+\.[\s\S]*?{[\s\S]*?}[\s\S]*?"  # contents that must have '{' and '}' symbols
        r"(?=<!--ID:\S+-->|^\d+\.|\Z|^---$)",  # card ends before ID, start of next card, end of string, end of section
        re.MULTILINE
    )
    _id_regex = re.compile(
        r'^<!--ID:(\S+)-->$',
        re.MULTILINE
    )
    _question_regex = re.compile(
        r'^\d+\.'  # question start 
        r'([\s\S]+?)'  # contents
        r'(?=^>|\Z)',  # answer start or end of string
        re.MULTILINE
    )
    _answer_regex = re.compile(
        r'(?:^>.*?(?:\n|$))+',
        re.MULTILINE
    )

    def __init__(
            self,
            file_path: Union[str, Path],
            default_deck: str,
    ):
        self._file_path = file_path
        self._default_deck = default_deck

    def collect_notes(self) -> List[Union[BasicNote, ClozeNote]]:
        """Get all notes from the file which path was passed to the Parser"""
        with open(self._file_path, mode='rt', encoding='utf-8') as f:
            file_string = f.read()

        question_sections = self._get_sections(file_string)

        notes = []
        for section in question_sections:
            notes.extend(self._get_notes_from_section(section))

        return notes

    def _get_notes_from_section(self, section: str) -> List[Union[BasicNote, ClozeNote]]:
        """Get all Notes from the section string"""
        tags = self._get_tags(section)
        deck_name = self._get_deck_name(section)

        note_strings = self.get_note_strings(section)

        # Create note objects
        notes = []
        for string in note_strings:
            anki_id = self.get_id(string)

            # we check in this order because is_cloze_note_str can match front/back note if it contains curly braces
            if self._is_basic_note_str(string):
                question = self.get_question(string)
                answer = self._get_cleaned_answer(string)

                notes.append(BasicNote(front_md=question,
                                       back_md=answer,
                                       tags=tags,
                                       deck_name=deck_name,
                                       anki_id=anki_id))
            elif self._is_cloze_note_str(string):
                text = self.get_question(string)

                notes.append(ClozeNote(text_md=text,
                                       tags=tags,
                                       deck_name=deck_name,
                                       anki_id=anki_id))
        return notes

    def _get_deck_name(self, section: str) -> str:
        """Get deck name specified for this section"""
        matches = re.findall(Parser._deck_name_regex, section)

        # If no deck name 
        if not matches:
            if not self._default_deck:
                raise ValueError(f"Couldn't find deck name in section:\n{section}")

            return self._default_deck

        if len(matches) > 1:
            raise ValueError(f'More than one deck name field in section:\n{section}')

        deck_name = matches[0].strip()
        if not deck_name:
            raise ValueError(f'Empty deck name field in section:\n{section}')

        return deck_name

    @classmethod
    def get_note_strings(cls, section: str) -> List[str]:
        """Get all strings of notes from section"""
        return re.findall(cls._all_notes_regex, section)

    @classmethod
    def get_id(cls, text: str) -> Optional[int]:
        """Get note's ID from text. Returns None if id wasn't found or if it is incorrect."""
        id_match = re.search(cls._id_regex, text)

        try:
            return int(id_match.group(1))
        except (ValueError, AttributeError):
            return None

    @classmethod
    def get_question(cls, text: str) -> str:
        """Get clean question string from text
         (without digit followed by period and trailing whitespace)"""
        question_match = re.search(cls._question_regex, text)

        question = question_match.group(1).strip()

        return question

    @classmethod
    def get_answer(cls, text: str) -> str:
        """Get answer string from text"""
        answer_match = re.search(cls._answer_regex, text)
        return answer_match.group()

    @classmethod
    def _get_sections(cls, file_contents: str) -> List[str]:
        """Get all sections (groups of notes) from the file string"""
        return re.findall(cls._section_regex, file_contents)

    @classmethod
    def _get_tags(cls, section: str) -> List[str]:
        """Get tags specified for this section"""
        matches = re.findall(cls._tags_regex, section)
        if not matches:
            return []

        if len(matches) > 1:
            raise ValueError(f'More than one tag field in section:\n{section}')

        tags = matches[0].strip().split()
        return tags

    @classmethod
    def _is_basic_note_str(cls, string: str) -> bool:
        """Check if note string contains basic note type"""
        match = re.search(cls._basic_note_regex, string)
        if not match:
            return False
        return True

    @classmethod
    def _is_cloze_note_str(cls, string: str) -> bool:
        """Check if note string contains cloze note type. Matches basic note type if it contains curly braces"""
        match = re.search(cls._cloze_note_regex, string)
        if not match:
            return False
        return True

    @classmethod
    def _get_cloze_note_strings(cls, section: str) -> List[str]:
        """Get all strings from section with only question and an (optional) ID"""
        return re.findall(cls._cloze_note_regex, section)

    @classmethod
    def _get_cleaned_answer(cls, text: str) -> str:
        """Get clean answer string from text (without '>' and trailing whitespace)"""
        # Remove '>' and first whitespace char after it (if there is any)
        lines = cls.get_answer(text).splitlines()
        cleaned_lines = []
        for line in lines:
            if len(line) > 1 and line[1].isspace():
                cleaned_lines.append(line[2:].rstrip())
            else:
                cleaned_lines.append(line[1:].rstrip())

        # Join lines differently: if inside code block '\n' else '\n\n'
        answer = ''
        inside_code_block = False
        for i, line in enumerate(cleaned_lines):
            delimiter = '\n\n'
            # If start or end of code block
            if line.find('```') != -1:
                if inside_code_block:
                    inside_code_block = False
                    delimiter = '\n'
                else:
                    inside_code_block = True
                    delimiter = '\n\n'
            elif inside_code_block:
                delimiter = '\n'

            # First line doesn't need delimiter
            if i == 0:
                delimiter = ''

            answer = answer + delimiter + line

        return answer



import os
import re
from typing import List, Dict, Optional, Union

from .anki_media import AnkiMedia
from .notes.basic_note import BasicNote
from .notes.cloze_note import ClozeNote


def handle_images_in(notes: List[Union[BasicNote, ClozeNote]], anki_media: AnkiMedia) -> None:
    """
    Copy images used in Notes fields to Anki Media folder and change source in their
    links to be just filename (for Anki to find them).

    Args:
        notes: Notes in which image links will be searched for and then updated
        anki_media: AnkiMedia object that will be used to copy images
    """
    # Find all unique image links in the notes
    image_links = _fetch_image_links(notes)

    # Copy images to Anki Media folder
    _copy_images_to(anki_media, list(image_links.keys()))

    # Update image links in notes
    _update_image_links_in_notes(image_links)


def _fetch_image_links(notes: List[Union[BasicNote, ClozeNote]]) -> Dict[str, List[BasicNote]]:
    """Get dictionary of image links with Note objects in which they are used.

    Args:
        notes: Notes in which image links will be searched
    Returns:
        Dictionary with image links as keys and a list of Note objects in which they are used as values
    """
    image_links = dict()
    image_regex = re.compile(r'!\[.*?]\(.*?\)')
    for note in notes:
        all_found_links = set()
        for field in note.get_raw_fields():
            all_found_links |= set(re.findall(image_regex, field))

        for link in all_found_links:
            if not image_links.get(link):
                image_links[link] = [note]
                continue
            image_links[link].append(note)

    return image_links


def _update_image_links_in_notes(image_links: Dict[str, List[Union[BasicNote, ClozeNote]]]) -> None:
    """Update image links in Notes. Updated text is stored in updated_front_md and updated_back_md of the Note.

    Args:
        image_links: dictionary with image links as keys and a list of Note objects in which they are used as values
    """
    image_path_regex = re.compile(r'(?<=\().+?(?=\))')
    for link, notes in image_links.items():
        image_filename = _get_filename_from(link)
        new_link = re.sub(image_path_regex, image_filename, link)

        for note in notes:
            # We use *updated* fields for replace func cause
            # we need to preserve previous replacements if there are multiple of them
            note.update_fields_with(lambda field: str.replace(field, link, new_link))


def _copy_images_to(anki_media: AnkiMedia, image_links: List[str]) -> None:
    """Copy images to Anki Media folder.

    Args:
        anki_media: AnkiMedia object that will be used to copy images
        image_links: list of markdown links to images
    Raises:
        FileNotFoundError: if path to image in markdown link is incorrect
        FileExistsError: if different file with the same name already exists in Anki Media folder
    """
    for link in image_links:
        abs_path = _get_abs_path_from(link)
        try:
            anki_media.copy_file_from(abs_path)
        except FileNotFoundError:
            raise FileNotFoundError(f'Cannot find image "{link}" on the path "{abs_path}"')


def _get_path_from(image_link: str) -> Optional[str]:
    """Get path used in markdown's image link.

    Args:
        image_link: markdown image link
    Returns:
        String with path used in markdown's image link
    """
    return re.search(r'(?<=\().+?(?=\))', image_link).group()


def _get_abs_path_from(image_link: str) -> Optional[str]:
    """Get absolute path to image from markdown's image link

    Args:
        image_link: markdown image link
    Returns:
        String with the absolute path to image
    """
    return os.path.realpath(
        _get_path_from(image_link)
    )


def _get_filename_from(image_link: str) -> Optional[str]:
    """Get image filename from markdown's image link
    Args:
        image_link: markdown image link
    Returns:
        String with the name of the image
    """
    return os.path.basename(
        _get_path_from(image_link)
    )



from typing import Iterable, Any, Callable, Dict, List

from .note import Note
from ..config import Config


class BasicNote(Note):
    """Front/Back note type"""

    def __init__(self, front_md: str, back_md: str, tags: Iterable[str], deck_name: str, anki_id: int = None):
        super().__init__(tags, deck_name, anki_id)
        self.raw_front_md = front_md
        self.raw_back_md = back_md
        self.updated_front_md = front_md  # With updated image links
        self.updated_back_md = back_md  # With updated image links
        self.front_html = None
        self.back_html = None

    def convert_fields_to_html(self, convert_func: Callable[[str], str]) -> None:
        """Convert note fields from markdown to html using provided function"""
        self.front_html = convert_func(self.updated_front_md)
        self.back_html = convert_func(self.updated_back_md)

    def update_fields_with(self, update_func: Callable[[str], str]) -> None:
        """Updates values of *updated* fields using provided function"""
        self.updated_front_md = update_func(self.updated_front_md)
        self.updated_back_md = update_func(self.updated_back_md)

    def get_search_field(self) -> str:
        """Get field (with html) that will be used for search in Anki"""
        return self.front_html

    def get_raw_fields(self) -> List[str]:
        """Get list of all raw (as in file) fields of this note"""
        return [self.raw_front_md, self.raw_back_md]

    def get_raw_question_field(self) -> str:
        """Get value of raw (as in file) question field"""
        return self.raw_front_md

    def get_html_fields(self, cfg: Config) -> Dict[str, str]:
        """Return dictionary with Anki field names as keys and html strings as values"""
        return {
            cfg.get_option_value('anki', 'front_field'): self.front_html,
            cfg.get_option_value('anki', 'back_field'): self.back_html
        }

    def get_note_info(self) -> str:
        """String used to display info about note in case of error"""
        front_shortened = self.shorten_text(self.raw_front_md)
        back_shortened = self.shorten_text(self.raw_back_md)
        info = 'Basic Note\n'
        info += '--------------------------------------------------\n'
        info += f'Front: {front_shortened}\n'
        info += f'Back: {back_shortened}\n'
        info += '--------------------------------------------------\n'
        return info

    @staticmethod
    def get_anki_note_type(cfg: Config) -> str:
        """Get name of Anki note type"""
        return cfg.get_option_value('anki', 'basic_type')

    def __eq__(self, other: Any) -> bool:
        if not super().__eq__(other):
            return False

        return (self.raw_front_md == other.raw_front_md and
                self.raw_back_md == other.raw_back_md)

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}(front_md={self.raw_front_md!r}, back_md={self.raw_back_md!r}, '
            f'tags={self.tags!r}, deck_name={self.deck_name!r}, anki_id={self.anki_id!r})'
        )



import re
from typing import Union, Tuple, AnyStr, Iterator, Match, Pattern

import mistune

from .notes.basic_note import BasicNote
from .notes.cloze_note import ClozeNote
from ..mistune_plugins.mathjax import plugin_mathjax, BLOCK_MATH, INLINE_MATH

MD = mistune.create_markdown(plugins=['strikethrough', 'footnotes', 'table', plugin_mathjax])

INLINE_CODE_REGEX = re.compile(
    r'`[\S\s]+?`',
    re.MULTILINE
)
BLOCK_CODE_REGEX = re.compile(
    r'```[\s\S]+?```',
    re.MULTILINE
)
CLOZE_DELETION_REGEX = re.compile(
    r'{{c\d+::[\s\S]*?}}|'
    r'{c?\d+::[\s\S]*?}|'
    r'{[\s\S]*?}',
    re.MULTILINE
)
ANKI_CLOZE_REGEX = re.compile(r'{{c\d+::[\s\S]*?}}', re.MULTILINE)
EXPLICIT_SHORT_CLOZE_REGEX = re.compile(r'{c?(\d+)::([\s\S]*?)}', re.MULTILINE)
IMPLICIT_SHORT_CLOZE_REGEX = re.compile(r'{([\s\S]*?)}', re.MULTILINE)

INLINE_CODE_PLACEHOLDER = 'INLINE_CODE_PLACEHOLDER'
BLOCK_CODE_PLACEHOLDER = 'BLOCK_CODE_PLACEHOLDER'
INLINE_MATH_PLACEHOLDER = 'INLINE_MATH_PLACEHOLDER'
BLOCK_MATH_PLACEHOLDER = 'BLOCK_MATH_PLACEHOLDER'


def convert_notes_to_html(notes: Iterator[Union[BasicNote, ClozeNote]]):
    """Convert note fields to html"""
    for note in notes:
        note.convert_fields_to_html(_convert_md_to_html)


def convert_cloze_deletions_to_anki_format(cloze_notes: Iterator[ClozeNote]):
    """Convert short syntax for cloze deletions to anki syntax"""
    for note in cloze_notes:
        # We substitute code and math inline/blocks with placeholders before searching cloze deletions
        # Block code should be removed before inline code
        block_code_matches, redacted_text = \
            _get_matches_and_updated_text(BLOCK_CODE_REGEX, BLOCK_CODE_PLACEHOLDER, note.updated_text_md)
        inline_code_matches, redacted_text = \
            _get_matches_and_updated_text(INLINE_CODE_REGEX, INLINE_CODE_PLACEHOLDER, redacted_text)

        # Block math should be removed before inline math
        block_math_matches, redacted_text = \
            _get_matches_and_updated_text(BLOCK_MATH, BLOCK_MATH_PLACEHOLDER, redacted_text)
        inline_math_matches, redacted_text = \
            _get_matches_and_updated_text(INLINE_MATH, INLINE_MATH_PLACEHOLDER, redacted_text)

        cloze_strings = CLOZE_DELETION_REGEX.findall(redacted_text)
        redacted_text = _replace_short_cloze_syntax(redacted_text, cloze_strings)

        redacted_text = _replace_placeholders_with_content(block_code_matches, BLOCK_CODE_PLACEHOLDER, redacted_text)
        redacted_text = _replace_placeholders_with_content(inline_code_matches, INLINE_CODE_PLACEHOLDER, redacted_text)
        redacted_text = _replace_placeholders_with_content(block_math_matches, BLOCK_MATH_PLACEHOLDER, redacted_text)
        redacted_text = _replace_placeholders_with_content(inline_math_matches, INLINE_MATH_PLACEHOLDER, redacted_text)

        note.updated_text_md = redacted_text


def _convert_md_to_html(text: str) -> str:
    # We delete '\n' before and after each html tag because Anki is rendering them as newlines
    return re.sub(r'\n?(<.+?>)\n?',
                  lambda tag_match: tag_match.group(1),
                  MD(text))


def _get_matches_and_updated_text(regex_pattern: Union[str, Pattern],
                                  placeholder: str,
                                  text: str) -> Tuple[Iterator[Match[AnyStr]], str]:
    """Search for matches with regex pattern and then replace same matches in text with placeholder"""
    matches = re.finditer(regex_pattern, text)
    new_text = re.sub(regex_pattern, placeholder, text)
    return matches, new_text


def _replace_placeholders_with_content(matches: Iterator[Match[AnyStr]], placeholder: str, text: str) -> str:
    """Replace placeholder occurrences with text from regex matches"""
    new_text = text
    for match in matches:
        new_text = new_text.replace(placeholder, match.group(0), 1)
    return new_text


def _replace_short_cloze_syntax(text: str, cloze_strings: Iterator[str]) -> str:
    """Replace short implicit and explicit cloze syntax with Anki cloze syntax in text"""
    new_text = text
    for i, cloze in enumerate(cloze_strings, start=1):
        if re.match(ANKI_CLOZE_REGEX, cloze):
            continue

        explicit_match = re.match(EXPLICIT_SHORT_CLOZE_REGEX, cloze)
        if explicit_match:
            explicit_index = explicit_match.group(1)
            content = explicit_match.group(2)
            new_cloze = '{{' + f'c{explicit_index}::{content}' + '}}'
            new_text = new_text.replace(cloze, new_cloze, 1)
            continue

        implicit_match = re.match(IMPLICIT_SHORT_CLOZE_REGEX, cloze)
        content = implicit_match.group(1)
        new_cloze = '{{' + f'c{i}::{content}' + '}}'
        new_text = new_text.replace(cloze, new_cloze, 1)

    return new_text



import os
import sys
from typing import Tuple, Set, List

import click
import requests

from . import __version__
from .models import highlighter, converter, img_handler
from .models.anki_api import AnkiApi
from .models.anki_media import AnkiMedia
from .models.config import Config
from .models.notes.basic_note import BasicNote
from .models.notes.cloze_note import ClozeNote
from .models.parser import Parser
from .models.writer import Writer

FILE_EXTENSIONS = ['.md', '.markdown']
CONFIG_PATH = f'{os.path.dirname(__file__)}/config.ini'

CONFIG = Config(CONFIG_PATH)


def get_notes_from_file(file_path: str) -> List[BasicNote]:
    click.echo('Getting cards from the file...')

    # We need to change working directory because images in file can have relative path
    os.chdir(os.path.dirname(file_path))

    default_deck = CONFIG.get_option_value('defaults', 'deck')
    parser = Parser(file_path, default_deck)
    notes = parser.collect_notes()

    return notes


def create_notes_from_file(file_path: str, anki_api: AnkiApi, anki_media: AnkiMedia) -> None:
    """Get all notes from file and send them to Anki"""
    click.echo(f'Starting to collect cards from "{os.path.basename(file_path)}"!')

    notes = get_notes_from_file(file_path)
    notes_num = len(notes)
    if notes_num == 0:
        click.echo(f"Cards weren't found!")
        return
    click.echo(f'Found {notes_num} cards!')

    converter.convert_cloze_deletions_to_anki_format(
        [note for note in notes if isinstance(note, ClozeNote)]
    )
    writer = Writer(file_path, notes)
    writer.update_cloze_notes()

    img_handler.handle_images_in(notes, anki_media)

    click.echo('Converting cards to the html...')
    converter.convert_notes_to_html(notes)

    click.echo('Synchronizing changes...')
    notes_with_id = [note for note in notes
                     if note.anki_id]
    anki_api.update_notes(notes_with_id)

    click.echo('Sending new cards...')
    notes_without_id = [note for note in notes
                        if not note.anki_id]
    anki_api.add_notes(notes_without_id)

    click.echo('Adding IDs to cards...')
    writer.update_note_ids()


def update_note_ids_in_file(file_path: str, anki_api: AnkiApi, anki_media: AnkiMedia):
    """Update IDs of notes in file by getting their IDs from Anki"""
    click.echo(f'Starting to update IDs of cards from "{os.path.basename(file_path)}"!')

    notes = get_notes_from_file(file_path)
    notes_num = len(notes)
    if notes_num == 0:
        click.echo(f"Cards weren't found!")
        return

    click.echo(f'Found {notes_num} cards!')
    converter.convert_cloze_deletions_to_anki_format(
        [note for note in notes if isinstance(note, ClozeNote)]
    )
    writer = Writer(file_path, notes)
    writer.update_cloze_notes()

    img_handler.handle_images_in(notes, anki_media)

    click.echo('Converting cards to the html...')
    converter.convert_notes_to_html(notes)

    click.echo('Getting card IDs from Anki...')
    anki_api.update_note_ids(notes)

    click.echo('Adding IDs to cards...')
    writer.update_note_ids()


def get_file_paths_from_directory(dir_path: str, search_recursive: bool) -> Set[str]:
    os.chdir(dir_path)

    paths_to_files = set()
    sub_directories = []
    for item in os.scandir():
        # Get markdown files
        extension = os.path.splitext(item.name)[1]
        if item.is_file() and extension in FILE_EXTENSIONS:
            paths_to_files.add(os.path.realpath(item.path))
        # And all directories
        elif search_recursive and item.is_dir():
            sub_directories.append(os.path.realpath(item.path))

    # Get all markdown files from each sub directory
    for directory in sub_directories:
        paths_to_files |= get_file_paths_from_directory(directory, search_recursive)
        os.chdir(dir_path)

    return paths_to_files


def get_paths_to_files(paths: Set[str], recursive: bool) -> Set[str]:
    paths_to_files = set()
    initial_directory = os.getcwd()
    for path in paths:
        full_path = os.path.realpath(path)

        if os.path.isdir(full_path):
            paths_to_files |= get_file_paths_from_directory(full_path, recursive)
            os.chdir(initial_directory)
            continue

        paths_to_files.add(full_path)

    return paths_to_files


def get_profile_from_user(profiles: List[str], message: str = None) -> str:
    if message:
        click.echo(message)
    profile = click.prompt('Enter the name of profile you want to use',
                           type=click.Choice(profiles))

    # Ask user to save profile as default
    if click.confirm('Save profile as default?'):
        CONFIG.update_option_value('defaults', 'profile', profile)

    return profile


def get_profile(prompt_user: bool, anki_api: AnkiApi) -> str:
    click.echo('Getting list of profiles from Anki...')
    profiles = anki_api.get_profiles()

    # If user has only one profile
    if len(profiles) == 1:
        return profiles[0]

    # If user used '-p' flag
    if prompt_user:
        return get_profile_from_user(profiles)

    click.echo('Getting profile from config...')
    profile = CONFIG.get_option_value('defaults', 'profile')
    if not profile:
        profile = get_profile_from_user(profiles, 'Default profile is not specified.')
    elif profile not in profiles:
        profile = get_profile_from_user(profiles, f'Incorrect profile name in config: {profile}')

    return profile


def check_anki_connection(anki_api: AnkiApi) -> None:
    click.echo("Attempting to connect to Anki...")
    if not anki_api.check_connection():
        click.secho("Couldn't connect to Anki. "
                    "Please, check that Anki is running and AnkiConnect plugin is installed.", fg='red')
        sys.exit(1)

    click.echo("Connected")


def reset_config_file(ctx, param, value) -> None:
    if not value or ctx.resilient_parsing:
        return

    CONFIG.reset()
    ctx.exit()


def list_config_options(ctx, param, value) -> None:
    if not value or ctx.resilient_parsing:
        return

    for entry in CONFIG.get_formatted_options():
        click.echo(entry)
    ctx.exit()


@click.group()
@click.version_option(version=__version__)
def cli() -> None:
    """Inka - command-line tool for adding flashcards from Markdown files to Anki.

        Documentation:\n
            https://github.com/lazy-void/inka/wiki

        Github:\n
            https://github.com/lazy-void/inka
    """
    pass


@cli.command()
@click.option('-l',
              '--list',
              'list_options',
              is_flag=True,
              callback=list_config_options,
              is_eager=True,
              help='List all options set in config file, along with their values.')
@click.option('-r',
              '--reset',
              is_flag=True,
              callback=reset_config_file,
              is_eager=True,
              help='Reset config file to default state.')
@click.argument('name',
                required=True)
@click.argument('value',
                required=False)
def config(list_options: bool, reset: bool, name: str, value: str) -> None:
    """Get and set inka's configuration options.

        NAME - config option name. VALUE - new value for config option.

        Examples:\n
            inka config defaults.profile "My Profile"
    """
    try:
        section, key = name.split('.')

        if not value:
            click.echo(CONFIG.get_option_value(section, key))
            sys.exit()

        CONFIG.update_option_value(section, key, value)
    except (KeyError, ValueError):
        click.secho('Incorrect name of a config entry!', fg='red')
        click.echo('To get list of all entries use "--list" flag.')
        sys.exit(1)


@cli.command()
@click.option('-r',
              '--recursive',
              is_flag=True,
              help='Search for files in subdirectories.')
@click.option('-p',
              '--prompt',
              is_flag=True,
              help='Show prompt for profile name even if config contains default profile.')
@click.option('-u',
              '--update-ids',
              'update_ids',
              is_flag=True,
              help='Update incorrect or absent IDs of cards from file(s) by searching for these cards in Anki.')
@click.argument('paths',
                metavar='[PATH]...',
                nargs=-1,
                type=click.Path(exists=True),
                required=False)
def collect(recursive: bool, prompt: bool, update_ids: bool, paths: Tuple[str]) -> None:
    """Get flashcards from files and add them to Anki. If flashcard already exists in Anki, the changes will be synced.

     If no PATH argument is passed, the program will use the path from config option 'defaults.folder'.

        [PATH]... - paths to files and/or directories

        Examples:\n
            inka collect path/to/cards.md - get cards from file\n
            inka collect path/to/folder - get cards from all files in folder
    """
    # If no path specified as an argument, search for default path in config
    paths = set(paths)
    if not paths:
        default_path = os.path.expanduser(CONFIG.get_option_value('defaults', 'folder'))
        if not default_path:
            click.secho('Default folder is not specified in the config!\n'
                        'You must pass the path to a file or folder as an argument.', fg='red')
            sys.exit(1)

        paths.add(default_path)

    anki_api = AnkiApi(CONFIG)

    # Check connection with Anki
    check_anki_connection(anki_api)

    # Get name of profile and select it in Anki
    profile = get_profile(prompt, anki_api)
    anki_api.select_profile(profile)

    anki_media = AnkiMedia(profile)
    try:
        highlighter.add_code_highlight_to(
            BasicNote,
            CONFIG.get_option_value('highlight', 'style'),
            anki_api,
            anki_media
        )
        highlighter.add_code_highlight_to(
            ClozeNote,
            CONFIG.get_option_value('highlight', 'style'),
            anki_api,
            anki_media
        )
    except KeyError:
        # Handle backward compatibility issues (config options were changed)
        click.secho('Your config file is corrupted. Please reset its state with the command "inka config --reset".',
                    fg='red')
        sys.exit(1)
    except requests.exceptions.RequestException as e:
        click.secho(f'Error while adding code highlight: "{e}"', fg='red')
        sys.exit(1)
    except ConnectionError as e:
        click.secho(str(e), fg='yellow')

    # Get paths to all files
    files = get_paths_to_files(paths, recursive)

    # Perform action on notes from each file
    initial_directory = os.getcwd()
    for file in files:
        try:
            if update_ids:
                update_note_ids_in_file(file, anki_api, anki_media)
            else:
                create_notes_from_file(file, anki_api, anki_media)

            os.chdir(initial_directory)
        except (OSError, ValueError, FileNotFoundError, FileExistsError) as e:
            click.secho(f'Error: "{e}"', fg='red')
            click.secho('Skipping file...', fg='red')
        except (requests.exceptions.RequestException, requests.exceptions.ConnectionError) as e:
            click.secho(f'Error while communicating with Anki: "{e}"', fg='red')
            sys.exit(1)

    click.echo('Finished!')



import filecmp
import os
import shutil
import sys
from pathlib import Path
from typing import Union

DEFAULT_ANKI_FOLDERS = {
    'win32': r'~\AppData\Roaming\Anki2',
    'linux': '~/.local/share/Anki2',
    'darwin': '~/Library/Application Support/Anki2'
}


class AnkiMedia:
    """Class for working with files in Anki Media folder"""

    def __init__(self, anki_profile: str) -> None:
        anki_folder_path = os.path.expanduser(DEFAULT_ANKI_FOLDERS[sys.platform])
        self._anki_media_path = f'{anki_folder_path}/{anki_profile}/collection.media'

    def exists(self, file_name: str) -> bool:
        """Check if file exists in Anki Media folder

        Args:
            file_name: name of the file
        Returns:
            bool: True if file exists and False otherwise
        """
        return os.path.exists(f'{self._anki_media_path}/{file_name}')

    def copy_file_from(self, file_path: Union[Path, str]) -> None:
        """Copy file to Anki Media folder.

        Args:
            file_path: path to file that will be copied
        Raises:
            FileNotFoundError: if file_path contains incorrect path
            FileExistsError: if different file with the same name already exists in Anki Media folder
        """
        file_name = os.path.basename(file_path)
        anki_file_path = f'{self._anki_media_path}/{file_name}'
        if self.exists(file_name):
            if filecmp.cmp(file_path, anki_file_path):
                return  # Skip if same file already exists

            raise FileExistsError(
                f'Different file with the same name "{file_name}" already exists in Anki Media folder.'
            )

        shutil.copyfile(file_path, anki_file_path)

    def create_file(self, file_name: str, file_content: str) -> None:
        """Create text file in Anki Media folder

        Args:
            file_name: name of the file
            file_content: content of the file
        Raises:
            FileExistsError: if file with the same name already exists in Anki Media folder
        """
        anki_file_path = f'{self._anki_media_path}/{file_name}'
        if self.exists(file_name):
            raise FileExistsError(f'File with the name "{file_name}" already exists in Anki Media folder.')

        with open(anki_file_path, 'wt', encoding='utf-8') as f:
            f.write(file_content)



from typing import Union, List, Any, Dict, Type

import click
import requests
from requests import RequestException

from .config import Config
from .notes.basic_note import BasicNote
from .notes.cloze_note import ClozeNote
from .notes.note import Note
from ..util import create_anki_search_query


class AnkiApi:
    """Class for working with server created by Anki Connect"""

    def __init__(self, cfg: Config):
        self._cfg = cfg
        self._api_url = f'http://localhost:{cfg.get_option_value("anki_connect", "port")}'
        self._change_tag = 'changed'
        self._delete_tag = 'delete'

    def check_connection(self) -> bool:
        """Check connection with Anki Connect plugin"""
        try:
            self._send_request('version')
        except requests.exceptions.ConnectionError:
            return False

        return True

    def get_profiles(self) -> List[str]:
        """Get list of user profiles from Anki"""
        return self._send_request('getProfiles')

    def select_profile(self, profile: str) -> None:
        """Select profile in Anki"""
        params = {'name': profile}
        self._send_request('loadProfile', **params)

    def add_notes(self, notes: List[Union[BasicNote, ClozeNote]]) -> None:
        """Add new notes to Anki"""
        # Create decks that doesn't exist
        decks = {note.deck_name for note in notes}
        for deck in decks:
            self._create_deck(deck)

        # Send notes to Anki
        for note in notes:
            note_params = self._create_note_params(note)
            try:
                note.anki_id = self._send_request('addNote', note=note_params)
            except RequestException as error:
                self._print_error_message(note, error)

    def update_note_ids(self, notes: List[Union[BasicNote, ClozeNote]]) -> None:
        """Update incorrect or absent IDs of notes"""
        # Handle None
        note_ids = [note.anki_id if note.anki_id else -1
                    for note in notes]

        notes_info = self._send_request('notesInfo', notes=note_ids)
        for i, note in enumerate(notes):
            # Don't update ID if note with this ID exists
            if notes_info[i]:
                continue

            search_field = note.get_search_field()
            query = create_anki_search_query(search_field)
            found_notes = self._send_request('findNotes', query=query)
            note.anki_id = found_notes[0] if found_notes else None

    def update_notes(self, notes: List[Union[BasicNote, ClozeNote]]) -> None:
        """Synchronize changes in notes with Anki"""
        # Get info about notes from Anki
        notes_info = self._send_request('notesInfo', notes=[note.anki_id for note in notes])

        for i, note in enumerate(notes):
            # If note wasn't found
            if not notes_info[i]:
                self._print_error_message(note, Exception(f'Card with ID {note.anki_id} was not found.'))
                continue

            # If note is staged for deletion
            if self._delete_tag in notes_info[i]['tags']:
                note.to_delete = True
                continue

            # If note is marked as changed
            # if self._change_tag in notes_info[i]['tags']:
            #     # Convert updated note fields to markdown
            #     note.front_html = notes_info[i]['fields'][self._front_field_name]['value']
            #     note.back_html = notes_info[i]['fields'][self._back_field_name]['value']
            #     # converter.convert_note_to_md(note)
            #
            #     # Mark note as changed
            #     note.changed = True
            #     continue

            try:
                # Push changes from file to Anki
                note_params = self._create_note_params(note)
                note_params['id'] = note.anki_id
                self._send_request('updateNoteFields', note=note_params)
            except RequestException as e:
                self._print_error_message(note, e)

    def delete_notes(self, notes: List[Union[BasicNote, ClozeNote]]) -> None:
        """Delete notes from Anki"""
        self._send_request('deleteNotes',
                           notes=[note.anki_id for note in notes])

    def remove_change_tag_from_notes(self, notes: List[Union[BasicNote, ClozeNote]]) -> None:
        """Remove the tag which marks note as changed from notes in Anki"""
        self._send_request('removeTags',
                           notes=[note.anki_id for note in notes],
                           tags=self._change_tag)

    def fetch_note_type_styling(self, note_type: Type[Note]) -> str:
        """Get styling of note type that is used to add notes"""
        return self._send_request('modelStyling', modelName=note_type.get_anki_note_type(self._cfg))['css']

    def update_note_type_styling(self, note_type: Type[Note], new_styles: str) -> None:
        """Update styling of note type that is used to add notes"""
        params = {
            'model': {
                'name': note_type.get_anki_note_type(self._cfg),
                'css': new_styles
            }
        }
        self._send_request('updateModelStyling', **params)

    def fetch_note_type_templates(self, note_type: Type[Note]) -> Dict[str, Dict[str, str]]:
        """Get templates of note type"""
        return self._send_request('modelTemplates', modelName=note_type.get_anki_note_type(self._cfg))

    def update_note_type_templates(self, note_type: Type[Note], templates: Dict[str, Dict[str, str]]) -> None:
        """Update note type templates"""
        params = {
            'model': {
                'name': note_type.get_anki_note_type(self._cfg),
                'templates': templates
            }
        }
        self._send_request('updateModelTemplates', **params)

    def _create_deck(self, deck: str) -> Any:
        """Create deck in Anki if it doesn't exist"""
        params = {'deck': deck}
        return self._send_request('createDeck', **params)

    def _create_note_params(self, note: Union[BasicNote, ClozeNote]) -> dict:
        """Create dict with params required to add note to Anki"""
        return {
            'deckName': note.deck_name,
            'modelName': note.get_anki_note_type(self._cfg),
            'fields': note.get_html_fields(self._cfg),
            'options': {
                'allowDuplicate': False,
                'duplicateScope': None,
                'duplicateScopeOptions': {
                    'checkChildren': False
                }
            },
            'tags': note.tags,
        }

    def _send_request(self, action: str, **params) -> Any:
        """Send request to Anki Connect and return result"""
        request_dict = self._create_request(action, **params)
        response = requests.post(self._api_url, json=request_dict).json()

        if len(response) != 2:
            raise RequestException('response has an unexpected number of fields')
        if 'error' not in response:
            raise RequestException('response is missing required error field')
        if 'result' not in response:
            raise RequestException('response is missing required result field')
        if response['error'] is not None:
            raise RequestException(response['error'])

        return response['result']

    @staticmethod
    def _create_request(action: str, **params) -> dict:
        """Create request dictionary"""
        return {'action': action, 'version': 6, 'params': params}

    @staticmethod
    def _print_error_message(note: Union[BasicNote, ClozeNote], error: Exception = None) -> None:
        click.echo()
        click.secho(note.get_note_info(), fg='red')

        if error is not None:
            click.secho(f'ERROR: "{error}"', fg='red')
        else:
            click.secho("ERROR: Can't create the card!", fg='red')
        input('Press Enter to continue...\n')



BLOCK_MATH = (
    r'(?<!\\)\$\$'
    r'([\s\S]*?)'
    r'(?<!\\)\$\$'
)

INLINE_MATH = (
    r'(?<![\\])\$'
    r'(.*?)'
    r'(?<![\\])\$'
)


def parse_inline_mathjax(inline, m, state):
    content = m.group(1)
    return 'mathjax_inline', content


def parse_block_mathjax(block, m, state):
    content = m.group(1)
    return 'mathjax_block', content


def render_html_inline_mathjax(content):
    return f'\\({content}\\)'


def render_html_block_mathjax(content):
    return f'\\[{content}\\]'


def plugin_mathjax(md):
    md.inline.register_rule('mathjax_block', BLOCK_MATH, parse_block_mathjax)
    md.inline.register_rule('mathjax_inline', INLINE_MATH, parse_inline_mathjax)

    md.inline.rules.append('mathjax_block')
    md.inline.rules.append('mathjax_inline')

    if md.renderer.NAME == 'html':
        md.renderer.register('mathjax_block', render_html_block_mathjax)
        md.renderer.register('mathjax_inline', render_html_inline_mathjax)



import re
from typing import Type

import requests
from requests import HTTPError

from .anki_api import AnkiApi
from .anki_media import AnkiMedia
from .notes.note import Note

HLJS_VERSION = "10.7.1"
BASE_URL = f"https://cdnjs.cloudflare.com/ajax/libs/highlight.js/{HLJS_VERSION}"

# Thanks u/arthurmilchior for the initial idea.
# https://www.reddit.com/r/Anki/comments/kzttw2/syntax_coloring_of_code_in_anki/
AUTOSTART_SCRIPT = """<script>
function color() {
    var selector = 'pre code';
    if (typeof hljs !== 'undefined') {
        codes = document.querySelectorAll(selector)
        codes.forEach(element => {
            hljs.highlightBlock(element);
        });
    } else {
        setTimeout(color, 50)
    }
}
color()
</script>"""


def add_code_highlight_to(
        note_type: Type[Note],
        style_name: str,
        anki_api: AnkiApi,
        anki_media: AnkiMedia
) -> None:
    """Add highlighting of code blocks to the current note type

    Args:
        note_type: the type of note to which the highlight will be added
        style_name: name of the highlight.js style to use
        anki_api: instance of AnkiApi used to modify note type
        anki_media: instance of AnkiMedia used to save highlight.js files
    Raises:
        ConnectionError: if something gone wrong during download of highlight.js style or script
    """
    try:
        _update_style_in(note_type, style_name, anki_api)
    except HTTPError as e:
        raise ConnectionError(f"Couldn't download styles for code highlighting. Reason: {e}")
    except requests.exceptions.ConnectionError:
        raise ConnectionError(f"Couldn't download styles for code highlighting. Check your internet connection.")

    try:
        _handle_highlighjs_files_for(note_type, anki_media, anki_api)
    except HTTPError as e:
        raise ConnectionError(f"Couldn't download highlight.js script for code highlighting. Reason: {e}")
    except requests.exceptions.ConnectionError:
        raise ConnectionError(
            f"Couldn't download highlight.js script for code highlighting. Check your internet connection.")


def _update_style_in(note_type: Type[Note], style_name: str, anki_api: AnkiApi) -> None:
    """Adds highlight.js style to the note type

    Args:
        note_type: the type of note to which styles will be added
        style_name: name of the highlight.js style to use
        anki_api: instance of AnkiApi used to modify note type
    Raises:
        ValueError: style_name is empty string or None
        HTTPError: something gone wrong during process of downloading style
    """
    if not style_name:
        raise ValueError("highlight style value in config must not be empty")

    current_styles = anki_api.fetch_note_type_styling(note_type)

    # Get current highlight.js style name and version
    start_regex = re.compile(r"/\*STYLE:(.+?) VERSION:(.+?)\*/")
    match = re.search(start_regex, current_styles)
    if match:
        current_highlight_style = match.group(1)
        version = match.group(2)
        if current_highlight_style == style_name and version == HLJS_VERSION:
            return

    # Download style from CDN (raise error if can't)
    url = f"{BASE_URL}/styles/{style_name}.min.css"
    response = requests.get(url)
    response.raise_for_status()
    new_highlight_style = response.text

    # Remove existing style
    highlight_regex = re.compile(r"\n/\*STYLE:(.+?) VERSION:(.+?)\*/[\s\S]+?/\*END\*/")
    current_styles = re.sub(highlight_regex, "", current_styles)

    # Add new style
    start = f"/*STYLE:{style_name} VERSION:{HLJS_VERSION}*/"
    end = "/*END*/"
    new_styles = f"{current_styles}\n{start}\n{new_highlight_style}\n{end}"

    anki_api.update_note_type_styling(note_type, new_styles)


def _handle_highlighjs_files_for(note_type: Type[Note], anki_media: AnkiMedia, anki_api: AnkiApi) -> None:
    """Adds highlight.js library and script that executes it to note type fields

    Args:
        note_type: the type of note to which links to files and autostart script will be added
        anki_api: instance of AnkiApi used to modify note type
        anki_media: instance of AnkiMedia used to save highlight.js files
    Raises:
        HTTPError: something gone wrong during process of downloading style
    """
    script_name = "_hl.pack.js"
    if not anki_media.exists(script_name):
        # Download script from CDN
        response = requests.get(f"{BASE_URL}/highlight.min.js")
        response.raise_for_status()
        script_content = response.text

        anki_media.create_file(script_name, script_content)

    # Get templates from note type
    templates = anki_api.fetch_note_type_templates(note_type)

    # Add link to script and script for automatic execution at the end of all fields of templates
    script_elements = f'<script src="{script_name}"></script>\n{AUTOSTART_SCRIPT}'
    new_templates = {}
    for template, fields in templates.items():
        new_templates[template] = {}
        for field, value in fields.items():
            new_templates[template][field] = value
            if value.find(script_elements) == -1:
                new_templates[template][field] += f"\n{script_elements}"

    # Don't update templates if nothing changed
    if templates == new_templates:
        return

    anki_api.update_note_type_templates(note_type, new_templates)



from setuptools import setup, find_packages

from src.inka import __version__

with open('README.md', mode='rt', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='inka',
    version=__version__,
    author='Kirill Salnikov',
    author_email='salnikov.k54@gmail.com',
    description='Command-line tool for adding flashcards from Markdown files to Anki',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/lazy-void/inka',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Intended Audience :: Education',
        'Topic :: Education :: Computer Aided Instruction (CAI)',
        'Topic :: Text Processing :: Markup :: Markdown',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only'
    ],
    license='GPLv3',
    keywords='anki, markdown, spaced-repetition',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.7',
    install_requires=[
        'mistune~=2.0.0rc1',
        'requests~=2.25.1',
        'click~=7.1.2'
    ],
    entry_points={
        'console_scripts': [
            'inka=inka.cli:cli'
        ]
    }
)



from abc import ABC, abstractmethod
from typing import Callable, Iterable, Any, Dict, List

from inka.models.config import Config


class Note(ABC):
    """Base class for all other note types"""

    def __init__(self, tags: Iterable[str], deck_name: str, anki_id: int = None):
        self.tags = tags
        self.deck_name = deck_name
        self.anki_id = anki_id
        self.changed = False  # Card was marked as changed in Anki
        self.to_delete = False  # Card was marked to be deleted in Anki

    @abstractmethod
    def convert_fields_to_html(self, convert_func: Callable[[str], str]) -> None:
        """Convert note fields from markdown to html using provided function"""
        pass

    @abstractmethod
    def update_fields_with(self, update_func: Callable[[str], str]) -> None:
        """Updates values of *updated* fields using provided function"""
        pass

    @abstractmethod
    def get_search_field(self) -> str:
        """Get field (with html) that will be used for search in Anki"""
        pass

    @abstractmethod
    def get_raw_fields(self) -> List[str]:
        """Get list of all raw (as in file) fields of this note"""
        pass

    @abstractmethod
    def get_raw_question_field(self) -> str:
        """Get value of raw (as in file) question field"""
        pass

    @abstractmethod
    def get_html_fields(self, cfg: Config) -> Dict[str, str]:
        """Return dictionary with Anki field names as keys and html strings as values"""
        pass

    @abstractmethod
    def get_note_info(self) -> str:
        """String used to display info about note in case of error"""
        pass

    @staticmethod
    @abstractmethod
    def get_anki_note_type(cfg: Config) -> str:
        """Get name of Anki note type"""
        pass

    @staticmethod
    def shorten_text(text: str) -> str:
        """Shorten text to the first 100 symbols"""
        return f'{text.strip()[:100]}...'

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return (self.tags == other.tags and
                self.deck_name == other.deck_name)
